import sys


def do_container(args):
    from janis_core.translations.janis import ToolTemplateType
    from janisdk.container.parse_help import from_container

    tooltype = ToolTemplateType.base
    if args.gatk4:
        tooltype = ToolTemplateType.gatk4

    outputdir = args.output

    tool, helpstr = from_container(
        container=args.container,
        basecommand=args.basecommand,
        helpcommand=args.help_str,
        containersoftware=args.container_tool,
        name=args.name,
        optionsmarker=args.options_marker,
        version=args.version,
        type=tooltype,
    )

    if args.printhelp:
        print(helpstr, file=sys.stderr)

    if args.printtool:
        print(tool, file=sys.stderr)

    if outputdir:
        from os import makedirs, path

        makedirs(outputdir, exist_ok=True)

        with open(path.join(outputdir, "base.py"), "w+") as f:
            f.write(tool)
        # with open(path.join(outputdir, "versions.py"), "w+") as f:
        #     f.write(toolversion)
        with open(path.join(outputdir, "__init__.py"), "w+"):
            pass
    else:
        print(tool, file=sys.stdout)
        # print(toolversion, file=sys.stdout)


def add_container_args(parser):
    parser.description = (
        "Attempts to parse the help (-h) guide of a tool and convert it into a "
        "CommandTool representation of Janis. The output of the tool is returned to stdout."
    )

    parser.add_argument("container", help="container to run")
    parser.add_argument("basecommand", help="The command of your tool", nargs="+")
    parser.add_argument(
        "-o",
        "--output",
        help="Directory to output base.py and versions.py to, otherwise these are both written to stdout",
    )

    annotations = parser.add_argument_group("Annotations")
    annotations.add_argument(
        "--version",
        help="Version of this tool, will try to get it from the container tag if not present",
        action="store_true",
    )
    annotations.add_argument(
        "--name", help="Name of tool, will default to name of command"
    )

    log_options = parser.add_argument_group("Logging options")

    log_options.add_argument(
        "--printhelp", help="Print the output of -h to stderr", action="store_true"
    )
    log_options.add_argument(
        "--printtool", help="Print the tool to stderr", action="store_true"
    )

    parser_info = parser.add_argument_group("Parsing options")
    parser_info.add_argument(
        "--options-marker",
        default="Options:",
        help="There's usually a header that separates the documentation from the parameters.",
    )
    parser_info.add_argument(
        "--help-str", help="String that your tool uses get the help guide", default="-h"
    )

    parser_info.add_argument("--container-tool", default="docker", choices=["docker"])

    extra = parser.add_argument_group("Extra options")
    extra.add_argument(
        "--gatk4", action="store_true", help="Use the GATK4 tool template"
    )



"""
Welcome to the source code for Janis! This module is mostly a placeholder and
used to bundle the different components of Janis (without crazy circular deps).
It also allows you to add your own tool sheds by including the:

    'janis.extension` entry-point.

Here's an amazing guide on entry-points:
    https://amir.rachum.com/blog/2017/07/28/python-entry-points/

Here are some of the tool sheds that are already built:

    - unix: janis-pipelines.unix (
    - bioinformatics: janis-pipelines.bioinformatics (https://github.com/PMCC-BioinformaticsCore/janis-bioinformatics)

There's also a workflow runner:
    - runner: janis-pipelines.runner (https://github.com/PMCC-BioinformaticsCore/janis-runner)

Janis is a framework creating specialised, simple workflow definitions that are
transpiled to Common Workflow Language or Workflow Definition Language.

_Proudly made on Planet Earth._

"""

from janis_core.__meta__ import __version__
from janis_core import *



import ast
import requests
from typing import List, Dict, Any, Optional
from janis_core.tool.test_suite_runner import ToolTestSuiteRunner
from janis_core.tool import test_helpers
from janis_core import Logger
from janis_assistant.engines.enginetypes import EngineType


class UpdateStatusOption:
    def __init__(self, url: str, token: str, method: Optional[str] = "patch"):
        self.url = url
        self.token = token
        self.method = method


class NotificationOption:
    def __init__(
        self, url: str, tool_name: str, test_case: str, test_id: Optional[str] = None
    ):
        self.url = url
        self.tool_name = tool_name
        self.test_case = test_case
        self.test_id = test_id


class TestCasesNotFound(Exception):
    pass


def run_test_case(
    tool_id: str,
    test_case: str,
    engine: EngineType,
    output: Optional[Dict] = None,
    config: str = None,
) -> Dict[str, Any]:
    tool = test_helpers.get_one_tool(tool_id)

    if not tool:
        raise Exception(f"Tool {tool_id} not found")

    runner = ToolTestSuiteRunner(tool, config=config)
    tests_to_run = [tc for tc in tool.tests() if tc.name.lower() == test_case.lower()]

    if not tests_to_run:
        raise Exception(f"Test case {test_case} not found")

    if len(tests_to_run) > 1:
        raise Exception(
            f"There is more than one test case with the same name {test_case}"
        )

    if output is not None:
        Logger.info(
            "Dryrun: validating test using provided output data without running the workflow"
        )

    failed = set()
    succeeded = set()
    execution_error = ""

    try:
        failed, succeeded, output = runner.run_one_test_case(
            t=tests_to_run[0], engine=engine, output=output
        )
    except Exception as e:
        execution_error = str(e)
    except SystemExit as e:
        execution_error = f"Workflow execution failed (exit code: {e.code})"

    return {
        "failed": list(failed),
        "succeeded": list(succeeded),
        "output": output,
        "execution_error": execution_error,
    }


def find_test_cases(tool_id: str):
    tool = test_helpers.get_one_tool(tool_id)

    if not tool:
        raise Exception(f"Tool {tool_id} not found")

    if tool.tests() is None:
        raise TestCasesNotFound(
            f"No test cases found for Tool {tool_id}. "
            f"You must implement the tests() function for this tool."
        )

    return [tc.name for tc in tool.tests()]


def update_status(result: Dict, option: UpdateStatusOption):
    Logger.info(f"Updating test status via {option.method} {option.url}")

    status = "test-failed"
    if not len(result["failed"]) and not result["execution_error"]:
        status = "test-succeeded"

    data = {"status": status, **result}

    headers = {"Authorization": f"Bearer {option.token}"}
    resp = requests.request(
        method=option.method, url=option.url, json=data, headers=headers
    )

    Logger.info("status updated")
    Logger.info(f"Response code {resp.status_code}")
    Logger.info(f"Response:\n{resp.text}")

    return resp.status_code, resp.text


def send_slack_notification(result: Dict, option: NotificationOption):
    Logger.info("sending notification to Slack")

    if len(result["failed"]) == 0 and not result["execution_error"]:
        failed = False
        status = "Test Succeeded"
        icon = ":white_check_mark:"
    else:
        failed = True
        status = "Test Failed"
        icon = ":x:"

    test_description = ""
    if option.test_id:
        test_description = f" *{option.test_id}*"

    summary_block = {
        "type": "section",
        "text": {
            "type": "mrkdwn",
            "text": f"{icon} {status}{test_description}: {option.tool_name} - {option.test_case}",
        },
    }

    blocks = [summary_block]

    if failed and result["failed"]:
        failed_expected_output = []

        for f in result["failed"]:
            failed_expected_output.append(f":black_small_square: {f}")

        failed_block = {
            "type": "section",
            "text": {"type": "mrkdwn", "text": "\n".join(failed_expected_output)},
        }

        blocks.append(failed_block)

    if result["execution_error"]:
        text = result["execution_error"].replace("\n", "<br />")
        execution_error_block = {
            "type": "section",
            "text": {"type": "mrkdwn", "text": f"{result['execution_error']}"},
        }

        blocks.append(execution_error_block)

    request = {"blocks": blocks}
    resp = requests.post(url=option.url, json=request)

    if resp.status_code == requests.codes.ok:
        Logger.info("Notification sent")
    else:
        Logger.warn("Failed to send slack notification")
        Logger.warn(f"{resp.status_code}: {resp.text}")

    return resp.status_code, resp.text


def cli_logging(result: Dict):
    name = result["test_case"]
    Logger.info(f"Test Case: {name}")
    Logger.info(f"Output: {result['output']}")

    if result["execution_error"]:
        Logger.critical(result["execution_error"])

    if len(result["succeeded"]) > 0:
        Logger.info(f"{len(result['succeeded'])} expected output PASSED")

        Logger.info("Succeeded expected output:")
        for s in result["succeeded"]:
            Logger.info(s)

    if len(result["failed"]) > 0:
        Logger.critical(f"{len(result['failed'])} expected output FAILED")

        Logger.critical("Failed expected output:")
        for f in result["failed"]:
            Logger.critical(f)

    if len(result["failed"]) == 0 and not result["execution_error"]:
        Logger.info(f"Test SUCCEEDED: {name}")
    else:
        Logger.critical(f"Test FAILED: {name}")


def add_runtest_args(parser):
    parser.add_argument("tool", help="Name of tool to test")

    parser.add_argument(
        "--test-case", help="Name of test case as listed in tool.tests()"
    )

    parser.add_argument("-e", "--engine", help="engine", default=EngineType.cromwell)

    parser.add_argument("-c", "--config", help="Path to janis config")

    parser.add_argument(
        "-o", "--output", help="Dry run test by providing a dictionary of output"
    )

    # For updating test-framework API endpoint
    parser.add_argument(
        "--test-manager-url", help="API endpoint to update run status in Test Manager"
    )

    parser.add_argument(
        "--test-manager-token", help="Authentication token for Test Manager API"
    )

    parser.add_argument(
        "--test-id",
        help="Test identification to be attached to the notification message",
    )

    parser.add_argument(
        "--slack-notification-url", help="Slack webhook to send notifications to"
    )


def execute(args):
    output = None
    if args.output:
        output = ast.literal_eval(args.output)

    try:
        available_test_cases = find_test_cases(args.tool)
        if args.test_case:
            if args.test_case not in available_test_cases:
                raise TestCasesNotFound(
                    f"Test case with name `{args.test_case}` NOT found."
                )
            test_cases = [args.test_case]
        else:
            test_cases = available_test_cases

    except Exception as e:
        Logger.critical("Unexpected error occurred when searching for test cases")
        Logger.critical(str(e))
        exit()

    for tc_name in test_cases:

        result = run_test_case(
            tool_id=args.tool,
            test_case=tc_name,
            engine=args.engine,
            output=output,
            config=args.config,
        )
        result["test_case"] = tc_name
        cli_logging(result)

        try:
            # send output to test framework API
            if args.test_manager_url and args.test_manager_token:
                option = UpdateStatusOption(
                    url=args.test_manager_url, token=args.test_manager_token
                )
                update_status(result, option)
        except Exception as e:
            Logger.warn(f"Failed to update test status to {args.test_manager_url}")

        try:
            # Send notification to Slack
            if args.slack_notification_url:
                option = NotificationOption(
                    url=args.slack_notification_url,
                    tool_name=args.tool,
                    test_case=tc_name,
                    test_id=args.test_id,
                )
                send_slack_notification(result=result, option=option)
        except Exception as e:
            Logger.warn(
                f"Failed to send notifications to Slack {args.slack_notification_url}"
            )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    add_runtest_args(parser)

    args = parser.parse_args()

    execute(args)



import importlib_metadata
from janis_core.utils.logger import Logger
from janis_core.toolbox.entrypoints import TOOLS

eps = importlib_metadata.entry_points().get(TOOLS, [])

for entrypoint in eps:
    try:
        m = entrypoint.load()
        for k, v in m.__dict__.items():
            if k.startswith("_"):
                continue
            globals()[k] = v
    except ImportError as e:
        Logger.critical(f"Couldn't import janis extension '{entrypoint.name}': {e}")
        continue



def add_fromcwl_args(parser):
    parser.description = (
        "Parse a CWL command line tool / workflow and write it as a Janis file"
    )

    parser.add_argument(
        "-o",
        "--output",
        help="Directory to output the workflow / tools to, otherwise this is written to stdout",
    )

    parser.add_argument("cwlfile", help="The path to the CWL file")

    parser.add_argument(
        "translation", default="janis", choices=["cwl", "wdl", "janis"], nargs="?"
    )

    return parser


def do_fromcwl(args):
    from janis_core import CWlParser, Logger

    Logger.info(f"Loading CWL file: {args.cwlfile}")
    tool = CWlParser.from_doc(args.cwlfile)

    Logger.info(f"Loaded {tool.type()}: {tool.versioned_id()}")

    translated = tool.translate(
        args.translation,
        to_console=args.output is None,
        to_disk=args.output is not None,
        export_path=args.output,
    )

    return translated



import unittest

from janisdk.cltconvert import get_string_repr, convert_commandtool
from janis_core import InputSelector, ToolOutput, String, File
from janis_unix.data_types.csv import Csv
from janis_unix.tools.echo import Echo


class DataTypes(unittest.TestCase):
    def test_string_nooptional(self):
        d = String()
        self.assertEqual("String()", get_string_repr(d))

    def test_string_optional(self):
        d = String(optional=True)
        self.assertEqual("String(optional=True)", get_string_repr(d))

    def test_file_nooptional(self):
        d = File()
        self.assertEqual("File()", get_string_repr(d))

    def test_file_optional(self):
        d = File(optional=True)
        self.assertEqual("File(optional=True)", get_string_repr(d))

    def test_csv_nooptional(self):
        d = Csv(optional=False)
        self.assertEqual("Csv()", get_string_repr(d))

    def test_csv_optional(self):
        d = Csv(optional=True)
        self.assertEqual("Csv(optional=True)", get_string_repr(d))


class TestInputSelector(unittest.TestCase):
    def test_inputselector_1param(self):
        i = InputSelector("filename")
        self.assertEqual(
            'InputSelector(input_to_select="filename")', get_string_repr(i)
        )


class TestToolOutput(unittest.TestCase):
    def test_tooloutput(self):
        o = ToolOutput("tag", String())
        self.assertEqual(
            'ToolOutput(tag="tag", output_type=String())', get_string_repr(o)
        )


class TestEcho(unittest.TestCase):
    def test_echo(self):
        tool = Echo()
        print(convert_commandtool(tool))


class TestDateTime(unittest.TestCase):
    def test_datetime(self):
        from datetime import datetime

        dt = datetime(1985, 11, 26, 1, 20)
        self.assertEqual("datetime(1985, 11, 26)", get_string_repr(dt))



from setuptools import setup, find_packages

# Version information is found in the __init__ file of `janis/`
DESCRIPTION = "Contains classes and helpers to build a workflow, and provide options to convert to CWL / WDL"

JANIS_CORE_VERSION = "v0.11.4"
JANIS_ASSISTANT_VERSION = "v0.11.7"
JANIS_UNIX_VERSION = "v0.11.0"
JANIS_BIOINFORMATICS_VERSION = "v0.11.1"
JANIS_PIPELINES_VERSION = "v0.11.3"
JANIS_TEMPLATES_VERSION = "v0.11.3"


######## SHOULDN'T NEED EDITS BELOW THIS LINE ########
fixed_core_version = f"janis-pipelines.core==" + JANIS_CORE_VERSION
fixed_assistant_version = f"janis-pipelines.runner==" + JANIS_ASSISTANT_VERSION
fixed_unix_version = f"janis-pipelines.unix==" + JANIS_UNIX_VERSION
fixed_bioinf_version = (
    f"janis-pipelines.bioinformatics==" + JANIS_BIOINFORMATICS_VERSION
)
fixed_pipes_version = f"janis-pipelines.pipelines==" + JANIS_PIPELINES_VERSION
fixed_templs_version = f"janis-pipelines.templates==" + JANIS_TEMPLATES_VERSION

with open("./README.md") as readme:
    long_description = readme.read()

vsn = {}
with open("./janis/__meta__.py") as fp:
    exec(fp.read(), vsn)
__version__ = vsn["__version__"]
githuburl = vsn["GITHUB_URL"]

modules = ["janis_assistant." + p for p in sorted(find_packages("./janis_assistant"))]


fixed_unix_version = f"janis-pipelines.unix==" + JANIS_UNIX_VERSION
setup(
    name="janis pipelines",
    version=__version__,
    description=DESCRIPTION,
    url=githuburl,
    author="Michael Franklin, Richard Lupat",
    author_email="michael.franklin@petermac.org",
    license="GNU",
    keywords=["pipelines", "bioinformatics"],
    packages=["janis", "janisdk"]
    + ["janis." + p for p in sorted(find_packages("./janis"))]
    + ["janisdk." + p for p in sorted(find_packages("./janisdk"))],
    install_requires=[
        fixed_core_version,
        fixed_assistant_version,
        fixed_unix_version,
        fixed_bioinf_version,
        fixed_pipes_version,
        fixed_templs_version,
    ],
    extras_require={
        "bioinformatics": [fixed_bioinf_version, fixed_pipes_version],
        "doc": ["docutils", "sphinx", "sphinx_rtd_theme", "recommonmark"],
        "ci": [
            "keyring==21.4.0",
            "setuptools",
            "wheel",
            "twine",
        ],
    },
    entry_points={"console_scripts": ["janisdk=janisdk.main:process_args"]},
    zip_safe=False,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)
